from typing import *
import operator
import re
from collections.abc import Iterable,Iterator,Generator,Collection


def test(a,b,cmp,cname=None):
    if cname is None: cname=cmp.__name__
    assert cmp(a,b),f"{cname}:\n{a}\n{b}"
def test_eq(a,b): test(a,b,operator.eq,'==')

def ifnone(a, b): return b if a is None else a
def setify(o): return o if isinstance(o,set) else set(listify(o))

def is_iter(o): return isinstance(o, (Iterable,Generator)) and getattr(o,'ndim',1)
def _is_array(x): return hasattr(x,'__array__') or hasattr(x,'iloc')
def listify(o):
    if o is None: return []
    if isinstance(o, list): return o
    if isinstance(o, str) or _is_array(o): return [o]
    if is_iter(o): return list(o)
    return [o]

def uniqueify(x, sort=False):
    res = list(OrderedDict.fromkeys(x).keys())
    if sort: res.sort()
    return res

class L():
    def __init__(self, items): self.items = listify(items)
    def __getitem__(self, idx):
        try: return self.items[idx]
        except TypeError:
            if isinstance(idx[0],bool):
                assert len(idx)==len(self) # bool mask
                return [o for m,o in zip(idx,self.items) if m]
            return [self.items[i] for i in idx]
    def __len__(self): return len(self.items)
    def __iter__(self): return iter(self.items)
    def __setitem__(self, i, o): self.items[i] = o
    def __delitem__(self, i): del(self.items[i])
    def __repr__(self):
        res = f'{self.__class__.__name__} ({len(self)} items)\n{self.items[:10]}'
        if len(self)>10: res = res[:-1]+ '...]'
        return res
    def append(self,o): return self.items.append(o)
    def remove(self,o): return self.items.remove(o)
    def unique(self): return L(dict.fromkeys(self).keys())
    def sort(self, key=None, reverse=False): return self.items.sort(key=key, reverse=reverse)
    def reverse(self ): return self.items.reverse()

_camel_re1 = re.compile('(.)([A-Z][a-z]+)')
_camel_re2 = re.compile('([a-z0-9])([A-Z])')
def camel2snake(name):
    s1 = re.sub(_camel_re1, r'\1_\2', name)
    return re.sub(_camel_re2, r'\1_\2', s1).lower()

def compose(x, funcs, *args, order_key='_order', **kwargs):
    key = lambda o: getattr(o, order_key, 0)
    for f in sorted(listify(funcs), key=key): x = f(x, **kwargs)
    return x



import sys, os
from pathlib import Path
import zipfile
import requests
import subprocess

Path.ls = lambda x: list(x.iterdir())

def get_dir_files(path, types=None):
    return [f for f in path.glob('**/*') if f.is_file() and (types==None or f.suffix[1:] in types)]

def get_dir_dirs(path):
    return [d for d in path.glob('**/*') if d.is_dir()]

def sh_run(cmd):
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    (stdout, stderr) = proc.communicate()
    return (proc.returncode, stdout.decode("utf-8") , stderr.decode("utf-8") )

def zipdir(dir, fname):
    zf = zipfile.ZipFile(fname, 'w', zipfile.ZIP_DEFLATED)
    for root, dirs, files in os.walk(dir):
        for file in files:
            zf.write(os.path.join(root, file))

def download_url(url, dest, overwrite=False, pbar=None, show_progress=True, chunk_size=1024*1024, timeout=4, retries=5):
    "Download `url` to `dest` unless it exists and not `overwrite`."
    # wget -c <url> && tar -zxvf <fname>
    if os.path.exists(dest) and not overwrite: return
    s = requests.Session()
    s.mount('http://',requests.adapters.HTTPAdapter(max_retries=retries))
    u = s.get(url, stream=True, timeout=timeout)
    try: file_size = int(u.headers["Content-Length"])
    except: show_progress = False

    with open(dest, 'wb') as f:
        nbytes = 0
        # if show_progress: pbar = progress_bar(range(file_size), auto_update=False, leave=False, parent=pbar)
        try:
            for chunk in u.iter_content(chunk_size=chunk_size):
                nbytes += len(chunk)
                if show_progress: pbar.update(nbytes)
                f.write(chunk)
        except requests.exceptions.ConnectionError as e:
            print(f'\n Download of {url} has failed after {retries} retries.\nError: {e}')
            return False
    return True

def download(url, fname=None, ext='.tgz'):
    "Download `url` to destination `fname`."
    if fname is None:
        fname = Path(f"{url.split('/')[-1]}{ext}")
    os.makedirs(fname.parent, exist_ok=True)
    if not fname.exists():
        print(f'Downloading {url}')
        if not download_url(f'{url}{ext}', fname):
            return None
    return fname



from datetime import datetime
import json
import uuid
import asyncio
import pika
from aio_pika import connect, connect_robust, Message, DeliveryMode, ExchangeType
from dl2050utils.env import Config_load
from dl2050utils.log import AppLog

# https://www.rabbitmq.com
# https://aio-pika.readthedocs.io/en/latest/index.html

JOB_CREATE = 0
JOB_START = 1
JOB_DONE = 2
JOB_NOTIFY = 3
JOB_DELIVER = 4
JOB_ERROR = 99

class MQ_SYNC():
    def __init__(self, log, db, qs, user='admin', passwd='password'):
        self.LOG, self.db, self.qs, self.user, self.passwd = log, db, qs, user, passwd
        
    def connect(self):
        try:
            mq_cred = pika.PlainCredentials(self.user, self.passwd)
            mq_conn = pika.BlockingConnection(pika.ConnectionParameters('mq', 5672, '/', mq_cred))
            self.ch = mq_conn.channel()
            for q in self.qs: self.ch.queue_declare(queue=q, durable=True, auto_delete=False)
        except Exception as e:
            self.LOG.log(4, 0, label='MQ', label2='startup', msg=str(e))
            return True
        self.LOG.log(2, 0, label='MQ', label2='startup', msg='CONNECTED')
        return False

    def startup(self): return self.connect()
    
    def server(self, queue, cb):
        if self.connect(): return True
        self.ch.basic_consume(queue=queue, auto_ack=False, on_message_callback=cb)
        self.ch.start_consuming()
        
    def publish(self, qname, uid, payload):
        job = {'uid': uid, 'payload': payload, 'jstatus': JOB_CREATE, 'eta': 0, 'ts_create': datetime.now()}
        jid = asyncio.run(self.db.insert('jobs', job, return_key='jid'))
        if jid is None: return None
        try:
            payload['uid'], payload['jid'] = uid, jid
            msg = Message(body=json.dumps(payload).encode(), delivery_mode=DeliveryMode.PERSISTENT)
            self.ch.basic_publish(exchange='', routing_key=qname, body=payload, properties=pika.BasicProperties(delivery_mode=2)) 
        except Exception as e:
            self.LOG.log(4, 0, label='MQ', label2='publish', msg=str(e))
            return None
        return jid

    def job_update(self, jid, status=None, eta=None, result=None):
        job = {'jid': jid}
        if status is not None:
            job['jstatus']=status
            if status==JOB_START: job['ts_start']=datetime.now()
            if status==JOB_DONE: job['ts_done']=datetime.now()
        if eta is not None: job['eta']=eta
        if result is not None: job['result']=result
        self.LOG.log(1, 0, label='MQ', label2='job_update', msg=job)
        res = asyncio.run(self.db.update('jobs', 'jid', job))
        if res is None or res==0: return True
        return False
    def job_start(self, jid, eta=None): return self.job_update(jid, status=JOB_START, eta=eta)
    def job_update_eta(self, jid, eta): return self.job_update(jid, eta=eta)
    def job_done(self, jid): return self.job_update(jid, status=JOB_DONE)
    def job_notify(self, jid): return self.job_update(jid, status=JOB_NOTIFY)
    def job_deliver(self, jid): return self.job_update(jid, status=JOB_DELIVER)
    def job_result(self, jid, result): return self.job_update(jid, result=result)
    def job_error(self, jid): return self.job_update(jid, status=JOB_ERROR, eta=0)
    def get_jobs(self, uid, status=None): return self.db.select('jobs', {'uid':uid, 'jstatus':status})
    def get_job(self, jid): return self.db.select_one('jobs', {'jid':jid})


class MQ():
    def __init__(self, log, db, qnames, user='admin', passwd='password'):
        self.LOG, self.db, self.qnames = log, db, qnames
        self.url = f'amqp://{user}:{passwd}@mq:5672/'
        # self.loop = asyncio.get_event_loop()
        # self.futures = {}
        
    async def connect(self):
        try:
            self.con = await connect_robust(self.url)
            self.ch = await self.con.channel()
            for qname in self.qnames: await self.ch.declare_queue(qname, durable=True, auto_delete=False)
        except Exception as e:
            self.LOG.log(4, 0, label='MQ', label2='connect', msg=str(e))
            return True
        self.LOG.log(2, 0, label='MQ', label2='connect', msg='CONNECTED')
        return False

    async def startup(self): return await self.connect()
    
    async def server(self, qname, cb):
        if await self.connect(): exit(1)
        queue = await self.ch.declare_queue(qname, durable=True, auto_delete=False)
        await queue.consume(cb, no_ack=False)
    
    async def publish(self, qname, uid, payload):
        job = {'uid': uid, 'payload': payload, 'jstatus': JOB_CREATE, 'eta': 0, 'ts_create': datetime.now()}
        jid = await self.db.insert('jobs', job, return_key='jid')
        if jid is None: return None
        try:
            payload['uid'], payload['jid'] = uid, jid
            msg = Message(body=json.dumps(payload).encode(), delivery_mode=DeliveryMode.PERSISTENT)
            await self.ch.default_exchange.publish(msg, routing_key=qname)
        except Exception as e:
            self.LOG.log(4, 0, label='MQ', label2='publish', msg=str(e))
            return None
        return jid
        
    async def job_update(self, jid, status=None, eta=None, result=None):
        job = {'jid': jid}
        if status is not None:
            job['jstatus']=status
            if status==JOB_START: job['ts_start']=datetime.now()
            if status==JOB_DONE: job['ts_done']=datetime.now()
        if eta is not None: job['eta']=eta
        if result is not None: job['result']=result
        self.LOG.log(1, 0, label='MQ', label2='job_update', msg=job)
        res = await self.db.update('jobs', 'jid', job)
        if res is None or res==0: return True
        return False
    async def job_start(self, jid, eta=None): return await self.job_update(jid, status=JOB_START, eta=eta)
    async def job_update_eta(self, jid, eta): return await self.job_update(jid, eta=eta)
    async def job_done(self, jid): return await self.job_update(jid, status=JOB_DONE)
    async def job_notify(self, jid): return await self.job_update(jid, status=JOB_NOTIFY)
    async def job_deliver(self, jid): return await self.job_update(jid, status=JOB_DELIVER)
    async def job_result(self, jid, result): return await self.job_update(jid, result=result)
    async def job_error(self, jid): return await self.job_update(jid, status=JOB_ERROR, eta=0)
    async def get_jobs(self, uid, status=None): return await self.db.select('jobs', {'uid':uid, 'jstatus':status})
    async def get_job(self, jid): return await self.db.select_one('jobs', {'jid':jid})



import os
from pathlib import Path
from dotenv import load_dotenv;
import yaml

def find_yml():
    config_path = os.getenv('CONFIG_PATH')
    if config_path is not None:
        p = Path(config_path) / '.Config.yml'
        if p.exists():
            return p
    dirs = ['.', '/app', '~']
    for e in dirs:
        p = Path(e) / '.Config.yml'
        if p.exists():
            return p
    return None

def Config_load():
    load_dotenv()
    fname = find_yml()
    if fname is None:
        return None, 'Config_load ERROR: Cant find .Config.yml'
    with open(str(fname), 'r') as stream:
        try:
            Config = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            return None, f'Config_load ERROR: .Config.yml parsing error (str({exc}))'
    return Config, None

def get_db_url(Config, with_db_name=False):
    if Config is None:
        return None
    if not 'db' in Config:
        return None
    if os.getenv('development') is None:
        if not 'host' in Config['db']:
            return None
        host = Config['db']['host']
    else:
        if not 'host_dev' in Config['db']:
            return None
        host = Config['db']['host_dev']
    if not 'port' in Config['db']:
        port = 3306
    else:
        port = Config['db']['port']
    if not 'name' in Config['db']:
        return None
    if not 'passwd' in Config['db']:
        return None
    passwd = Config['db']['passwd']
    url = f'mysql+pymysql://root:{passwd}@{host}:{port}'
    if with_db_name:
        name = Config['db']['name']
        url = f'{url}/{name}'
    return url

def get_rest_url(Config):
    if Config is None:
        return None
    if not 'rest' in Config:
        return None
    if os.getenv('development') is None:
        if not 'host' in Config['rest']:
            return None
        host = Config['rest']['host']
    else:
        if not 'host_dev' in Config['rest']:
            return None
        host = Config['rest']['host_dev']
    if not 'port' in Config['rest']:
        return None
    port = Config['rest']['port']
    url = f'{host}:{port}'
    return url


import asyncpg
import datetime
import json
from dl2050utils.core import *

# https://www.psycopg.org
# https://magicstack.github.io/asyncpg/current/index.html


TBLS_EXCLUDE = ['refs', 'trans', 'enums', 'information_schema_catalog_name', 'check_constraint_routine_usage', 'applicable_roles',
                'administrable_role_authorizations','collation_character_set_applicability', 'attributes', 'check_constraints',
                'character_sets', 'collations', 'column_domain_usage', 'column_column_usage', 'column_privileges',
                'column_udt_usage', 'columns', 'constraint_column_usage', 'schemata', 'constraint_table_usage',
                'domain_constraints', 'sql_packages', 'domain_udt_usage', 'sequences', 'domains', 'enabled_roles',
                'key_column_usage', 'parameters', 'referential_constraints', 'sql_features', 'role_column_grants',
                'routine_privileges', 'role_routine_grants', 'routines', 'sql_implementation_info', 'sql_parts',
                'sql_languages', 'sql_sizing', 'sql_sizing_profiles', 'table_constraints', 'table_privileges',
                'role_table_grants', 'views', 'tables', 'transforms', 'triggered_update_columns', '_pg_foreign_servers',
                'triggers', 'data_type_privileges', 'udt_privileges', 'role_udt_grants', 'usage_privileges', 'element_types',
                'role_usage_grants', 'user_defined_types', '_pg_foreign_table_columns', 'view_column_usage', 'view_routine_usage',
                'view_table_usage', 'foreign_server_options', 'column_options', '_pg_foreign_data_wrappers',
                'foreign_data_wrapper_options', 'foreign_tables', 'foreign_data_wrappers', 'foreign_servers',
                'foreign_table_options', 'user_mappings', 'user_mapping_options']

DRILLS = {
    'tbls': {'canal': ['zona', 'subzona', 'agente']}
}

TYPES = {
    'character varying': 'S',
    'character': 'S',
    'integer': 'I',
    'real': 'P',
    'double precision': 'F',
    'money': 'C',
    'date': 'D',
    'timestamp with time zone': 'T',
    'boolean': 'B',
    'ARRAY': 'ARRAY',
    'json': 'JSON',
    'USER-DEFINED': 'ENUM'
}

Q_DROP = """
    DROP TABLE flds;
    DROP TABLE enums;
    DROP TABLE refs;
    DROP TABLE tbls;
    DROP TYPE aligns;
"""

Q_CREATE = """
    CREATE TYPE aligns AS ENUM ('Left', 'Center', 'Right');
    
    CREATE TABLE tbls (
        tbl varchar(64) primary key,
        name varchar(64),
        keys TEXT [],
        descf varchar(64),
        drills json
    );

    CREATE TABLE flds (
        tbl varchar(64) NOT NULL references tbls(tbl),
        fld varchar(64) NOT NULL,
        name varchar(64),
        type varchar(64),
        size char(16),
        frmt char(16),
        ronly boolean,
        align aligns,
        PRIMARY KEY(tbl, fld)
    );
    
    CREATE TABLE refs (
        tbl varchar(64) NOT NULL references tbls(tbl),
        fld varchar(64) NOT NULL,
        tbl_ref varchar(64) NOT NULL references tbls(tbl),
        fld_ref varchar(64) NOT NULL
    );
    
    CREATE TABLE enums (
        name varchar(64) NOT NULL,
        values TEXT []
    );
    
"""

Q_TBLS = """
    SELECT table_name FROM information_schema.tables;
"""

Q_FLDS_ = f"""
    SELECT table_name as tbl, column_name as fld, column_name as name, data_type as type, character_maximum_length as size
    FROM INFORMATION_SCHEMA.COLUMNS
    WHERE table_name IN 
"""

Q_KEYS = """
    SELECT tc.table_name as tbl, c.column_name as fld
    FROM information_schema.table_constraints tc 
    JOIN information_schema.constraint_column_usage AS ccu USING (constraint_schema, constraint_name) 
    JOIN information_schema.columns AS c ON c.table_schema = tc.constraint_schema
        AND tc.table_name = c.table_name AND ccu.column_name = c.column_name
    WHERE constraint_type = 'PRIMARY KEY';
"""

Q_REFS = """
    SELECT
        tc.table_name as tbl, 
        kcu.column_name as fld, 
        ccu.table_name AS tbl_ref,
        ccu.column_name AS fld_ref
    FROM 
        information_schema.table_constraints AS tc 
        JOIN information_schema.key_column_usage AS kcu
          ON tc.constraint_name = kcu.constraint_name
          AND tc.table_schema = kcu.table_schema
        JOIN information_schema.constraint_column_usage AS ccu
          ON ccu.constraint_name = tc.constraint_name
          AND ccu.table_schema = tc.table_schema
    WHERE tc.constraint_type = 'FOREIGN KEY' and tc.constraint_name ='flds_tbl_fkey';
"""

Q_ENUMS = """
    select t.typname as ename,  
           e.enumlabel as evalue
    from pg_type t 
       join pg_enum e on t.oid = e.enumtypid  
       join pg_catalog.pg_namespace n ON n.oid = t.typnamespace;
"""

async def build_meta(db):
    for q in Q_DROP.split(";"): await db.query(q)
    for q in Q_CREATE.split(";"): await db.query(q)
    
    tbls = await db.query(Q_TBLS)
    tbls_list = [e['table_name'] for e in tbls
            if e['table_name'] not in TBLS_EXCLUDE and not e['table_name'].startswith('pg') and not e['table_name'].startswith('_pg')]
    keys = await db.query(Q_KEYS)
    tbls = [{'tbl': e, 'name': e, 'keys': [e1['fld'] for e1 in keys if e1['tbl']==e], 'drills': DRILLS[e] if e in DRILLS else {}}
            for e in tbls_list]
    await db.insert_rows('tbls', tbls, delete=False)
    
    TBLS_LIST = ', '.join([f'\'{e}\'' for e in tbls_list])
    Q_FLDS = f'{Q_FLDS_} ({TBLS_LIST});'
    flds = await db.query(Q_FLDS)
    for fld in flds:
        fld['name'] = fld['name'].title()
        fld['type'] = TYPES[fld['type']]
        fld['align'] = 'Left' if fld['type'] in ['S', 'D', 'T'] else 'Right'
    await db.insert_rows('flds', flds, delete=True)
        
    enums = await db.query(Q_ENUMS)
    enums = [{'name': e, 'values': [e1['evalue'] for e1 in enums if e1['ename']==e]} for e in set([e['ename'] for e in enums])]
    await db.insert_rows('enums', enums, delete=True)
    
    refs = await db.query(Q_REFS)
    await db.insert_rows('refs', refs, delete=True)


async def update_meta(db, meta):
    for tbl in meta['tables']:
        if 'fields' not in meta['tables'][tbl]: continue
        for fld in meta['tables'][tbl]['fields']:
            d = {'tbl': tbl, 'fld': fld}
            for attr in meta['tables'][tbl]['fields'][fld]:
                d[attr] = meta['tables'][tbl]['fields'][fld][attr]
            res = await db.update('flds', ['tbl', 'fld'], d)


async def get_meta(db):
    tbls = await db.select('tbls')
    if tbls is None: return None
    flds = await db.select('flds')
    if flds is None: return None
    meta = {'tables': {e['tbl']: {'table': e, 'fields': {e1['fld']: e1 for e1 in flds if e1['tbl']==e['tbl']}} for e in tbls}}
    return meta


def fix_types(d):
    for k in d.keys():
        if isinstance(d[k], str): d[k] = d[k].strip()
        if isinstance(d[k], datetime.date):
            d[k] = d[k].strftime("%Y-%m-%d %H:%M:%S") if isinstance(d[k], datetime.datetime) else d[k].strftime("%Y-%m-%d")
    return d

def strip(e):
    if type(e) != str: return e
    e = e.replace('\'', '')
    e.replace('\"', '')
    e.replace('\n', ' ')
    return e

def get_repr(e):
    if type(e)==list:
        items = [f'"{str(e1)}"' for e1 in e]
        return f"'{{{' ,'.join(items)}}}'"
    if type(e)==dict: return f"'{json.dumps(e)}'"
    return f"'{strip(e)}'"


class DB():
    def __init__(self, cfg, log):
        self.cfg, self.LOG = cfg, log
        try:
            passwd = self.cfg["db"]["passwd"]
        except Exception as e:
            passwd = 'rootroot'
        self.url = f'postgres://postgres:{passwd}@db:5432/postgres'

    async def startup(self, min_size=5, max_size=20):
        try:
            self.pool = await asyncpg.create_pool(self.url, min_size=min_size, max_size=max_size)
        except Exception as e:
            self.LOG.log(4, 0, label='DBPG', label2='startup', msg=str(e))
            return True
        self.LOG.log(2, 0, label='DBPG', label2='startup', msg='CONNECTED (POOL)')
        return False
    
    async def build_meta(self): return await build_meta(self)
    async def update_meta(self, meta): return await update_meta(self, meta)
    async def get_meta(self): return await get_meta(self)

    def shutdown(self):
        self.pool.terminate()  # self.pool.close()
        self.LOG.log(2, 0, label='DBPG', label2='shutdown', msg='DISCONNECTED')
        return False
    
    async def execute(self, q):
        con = await self.pool.acquire()
        try:
            res = await con.execute(q)
        except Exception as e:
            self.LOG.log(4, 0, label='DBPG', label2='execute', msg={'error_msg': str(e), 'query': q})
            return None
        finally:
            await self.pool.release(con)
        # self.LOG.log(1, 0, label='DBPG', label2='DEBUG', msg=q)
        return res
        
    async def query(self, q, one=False):
        con = await self.pool.acquire()
        try:
            if(one):
                res = await con.fetchrow(q)
                if res is None: return None
            else:
                res = await con.fetch(q)
        except Exception as e:
            self.LOG.log(4, 0, label='DB', label2='query', msg=str(e))
            return None
        finally:
            await self.pool.release(con)
        if(one): return fix_types(dict(res))
        # self.LOG.log(1, 0, label='DBPG', label2='DEBUG', msg=q)
        return [fix_types(dict(row)) for row in res]
    
    async def get_trows(self, tbl):
        q = f"SELECT COUNT(*) FROM {tbl};"
        con = await self.pool.acquire()
        try:
            res = await con.fetchrow(q)
            return dict(res)['count']
        except Exception as e:
            self.LOG.log(4, 0, label='DB', label2='get_trows', msg=str(e))
            return 0

    async def select(self, tbl, d=None, cols='*', order=None, ascending=True, offset=None, limit=None, one=False):
        if cols!='*': cols = ', '.join(cols)
        q = f"SELECT {cols}, count(*) OVER() AS nrows FROM {tbl}"
        if d is not None and len(d):
            filters = ' AND '.join([f"{k}='{d[k]}'" for k in d if d[k] is not None])
            q += f" WHERE {filters}"            
        if order is not None: q += f" ORDER BY {order} " + ("ASC" if ascending else "DESC")
        if offset is not None: q += f" OFFSET {offset}"
        if limit is not None: q += f" LIMIT {limit}"
        q += ';'
        # self.LOG.log(1, 0, label='DBPG', label2='select', msg=q)
        return await self.query(q, one=one)

    async def select_one(self, tbl, d): return await self.select(tbl, d=d, one=True)
    
    async def insert(self, tbl, d, return_key=None):
        q = f"INSERT INTO {tbl} ("
        for k in d.keys(): q += f"{k}, "
        q = q[:-2] + ") VALUES ("
        for k in d.keys(): q = q + get_repr(d[k]) + ', '
        q = q[:-2] + ")"
        # self.LOG.log(1, 0, label='DBPG', label2='insert',  msg=q)
        if return_key is not None:
            q += f" returning {return_key};"
            res = await self.query(q)
            if res is not None and len(res): return res[0][return_key]
            return None
        return await self.execute(q)
    
    async def insert_rows(self, tbl, rows, delete=False):
        if delete:
            res = await self.execute(f'DELETE FROM {tbl}')
            if res is None: return res
        for row in rows:
            res = await self.insert(tbl, row)
            if res is None: return res
        return len(rows)

    async def update(self, tbl, ks, d):
        ks = listify(ks)
        for k in ks:
            if not k in d:
                self.LOG.log(4, 0, label='DBPG', label2='update', msg=f'key error: {k}')
                return None
        setvars = ', '.join([f"{k}={get_repr(d[k])}" for k in d if k not in ks+['nrows'] and d[k] is not None])
        q = f"UPDATE {tbl} SET {setvars}"
        filters = ' AND '.join([f"{k}='{d[k]}'" for k in ks if d[k] is not None])
        q += f" WHERE {filters};"
        # self.LOG.log(1, 0, label='DBPG', label2='update',  msg=q)
        res = await self.execute(q)
        if res is None: return None
        return int(res[7:])

    async def delete(self, tbl, k, v):
        q = f"DELETE FROM {tbl}"
        if k is not None and v is not None:
            q += f" WHERE {k}='{v}'"
        q += ';'
        return await self.execute(q)
    
    async def headers(self, tbl, k, v):
        pass


if __name__ == "__main__":
    print(L([1,2,3]))


import sys, logging, time
from datetime import datetime

LOG_FORMAT = '\
%(name)s  %(levelname)s - %(asctime)s - %(product)s:%(client)s:%(instance)s:%(service)s:%(process)s - \
%(label)s - %(label2)s - %(duration).3f - %(message)s'

def parse_msg(d, level=0):
    if d is None: return 'null'
    if type(d) in [int, float]: return str(d)
    if isinstance(d, str): return d.replace("\n", " ")
    if isinstance(d, datetime): return d.strftime("%Y-%m-%Y %H:%M:%S")
    if type(d) is list: return f'[#{len(d)}]: [' + ', '.join(f'{v}' for v in d[:10]) + ']'
    if type(d) is not dict: return f'OBJECT-{type(d)}'
    s = '{'
    for key, val in d.items():
        if isinstance(d[key], dict):
            s += f'{key}={parse_msg(d[key],level+1)[:-1]}'
            continue
        if type(d[key]) is list:
            s += f'{key}[#{len(d[key])}]: '
            s += '[' + ', '.join(f'{v}' for v in d[key][:10]) + ']'
        else:
            s += f'{key}={val}'
        s += ','
    s = s[:-1]+'}, '
    if level==0 and s[-1]==' ': s=s[:-2]
    return s

class AppLog():
    def __init__(self, cfg=None, product='_PRODUCT_', client='_CLIENT_', instance='_INSTANCE_', service='_SERVICE_'):
        log_format = logging.Formatter(LOG_FORMAT)
        logger = logging.getLogger('APPLOG')
        self.logger = logger
        handler = logging.StreamHandler()
        # handler.flush = sys.stdout.flush
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(log_format)
        self.handler = handler
        self.LOG_LEVELS = {
            1: logger.debug,
            2: logger.info,
            3: logger.warning,
            4: logger.error,
            5: logger.critical
        }
        self.msg_prefix = {}
        self.msg_prefix['service'] = service
        if cfg is not None:
            try:
                self.msg_prefix['product']  = cfg['app']['product']
                self.msg_prefix['client'] = cfg['app']['client']
                self.msg_prefix['instance'] = cfg['app']['instance']
                return
            except Exception as e:
                pass
        self.msg_prefix['product'] = product
        self.msg_prefix['client'] = client
        self.msg_prefix['instance'] = instance
                    
    def setup(self):
        for key in logging.Logger.manager.loggerDict:
            logging.getLogger(key).setLevel(logging.CRITICAL)
        self.logger.addHandler(self.handler)
        self.logger.setLevel(logging.DEBUG)
        self.logger.propagate = False

            
    def log(self, level, ts0=0, label=None, label2=None, msg=''):
        if not isinstance(level, int) or level<1 or level>5: return
        log_f = self.LOG_LEVELS[level]
        log_msg = self.msg_prefix
        log_msg['label'] = label
        log_msg['label2'] = label2
        log_msg['duration'] = ts0
        s = parse_msg(msg)
        log_f(s, extra=log_msg)


import os, re
from os.path import basename
import json
import socket
import smtplib, ssl
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email import encoders
from multiprocessing import Process
import boto3


default_address_book = {
    'jneto': {'email': 'joao.filipe.neto@gmail.com', 'phone': '+351966221506'},
    'jranito': {'email': 'joao.vasco.ranito@gmail.com', 'phone': '+351966221505'},
}

def read_address_book():
    try:
        with open('./.addressbook') as f:
            d = json.loads(f.read())
    except IOError:
        return None
    return d


#https://realpython.com/python-send-email/
def send_mail_async(email_id, email_passwd, to, subject=None, message=None, html=None, files=[]):
    p = Process(target=send_mail, args=(email_id, email_passwd, to, subject, message, html, files), daemon=True)
    p.start()

def send_mail(email_id, email_passwd, to, subject=None, message=None, html=None, files=[]):
    smtp_server = 'smtp.gmail.com'
    smtp_port = 465
    msg = MIMEMultipart('mixed')
    msg['Subject'] = subject if subject is not None else ''
    if not type(to) is list:
        to = [to]
    msg['To'] = ', '.join(to)
    msg['From'] = email_id
    body = MIMEText(message, "plain") if html is None else MIMEText(html, 'html')
    msg.attach(body)
    for file in files:
        part = MIMEBase('application', 'octet-stream')
        part.set_payload(open(file, "rb").read())
        encoders.encode_base64(part)
        arg = 'attachment; filename="' + basename(file) + '"'
        part.add_header('Content-Disposition', arg)
        msg.attach(part)
    context = ssl.create_default_context()
    with smtplib.SMTP_SSL(smtp_server, smtp_port, context=context) as server:
        try:
            res = server.login(email_id, email_passwd)
            res = server.sendmail(email_id, to, msg.as_string())
        except Exception as e:
            return f'Mail sender ERROR: {str(e)}'
    return None


def send_sms_aws(sms_id, sms_passwd, to, msg):
    MessageAttributes={'AWS.SNS.SMS.SenderID': {'DataType': 'String','StringValue': 'DLOPS'}}
    client = boto3.client('sns', aws_access_key_id=sms_id, aws_secret_access_key=sms_passwd, region_name="eu-west-1")
    client.publish(PhoneNumber=to, Message=msg, MessageAttributes=MessageAttributes)
    return None

#from twilio.rest import Client
#def send_sms_twilio(sms_id, sms_passwd, to, msg):
#    #account_sid, auth_token
#    client = Client(sms_id, sms_passwd)
#    message = client.messages.create(to=to, from_='DLbot', body=msg)
#    #print(message.sid)
#    return None

def send_sms(sms_id, sms_passwd, to, msg):
    return send_sms_aws(sms_id, sms_passwd, to, msg)


class Notify():
    def __init__(self, cfg=None, email_id=None, email_passwd=None, sms_id=None, sms_passwd=None, address_book=None):
        self.address_book = address_book if address_book is not None else default_address_book
        if cfg is not None:
            try:
                email_id = cfg['email']['user']
                email_passwd = cfg['email']['passwd']
                sms_id = cfg['aws']['aws_access_key_id']
                sms_passwd = cfg['aws']['aws_secret_access_key']
            except Exception as e:
                print(f'Config ERROR: cant find variable: {e}')
        self.email_id = email_id
        self.email_passwd = email_passwd
        self.sms_id = sms_id
        self.sms_passwd = sms_passwd
        
    def __call__(self, who, how='mail', subject=None, message=None, html=None, files=[]):
        if who is None or who not in self.address_book:
            return 'Destination not found in address book'
        if how=='email' or how=='email_async':
            if self.email_id is None or self.email_passwd is None:
                return 'email credentials not defined'
            if 'email' not in self.address_book[who]:
                return f'email address not found for {who}'
            to = self.address_book[who]['email']
        if how=='sms':
            if self.sms_id is None or self.sms_passwd is None:
                return 'sms credentials not defined'
            if 'phone' not in self.address_book[who]:
                return f'phone number not found for {who}'
            to = self.address_book[who]['phone']
        if message is None:
            return 'Nothing to notify'
        if how == 'email':
            return send_mail(self.email_id, self.email_passwd, to, subject=subject, message=message,
                             html=html, files=files)
        if how == 'email_async':
            return send_mail_async(self.email_id, self.email_passwd, to, subject=subject, message=message,
                                   html=html, files=files)
        if how == 'sms':
            return send_sms(self.sms_id, self.sms_passwd, to, message)
        return 'Invalid method, options are email, email_async or sms'
    
    def send_mail_async(self, to, subject=None, message=None, html=None, files=[]):
        send_mail_async(self.email_id, self.email_passwd, to, subject=subject, message=message,
                             html=html, files=files)
    
    def send_mail(self, to, subject=None, message=None, html=None, files=[]):
        return send_mail(self.email_id, self.email_passwd, to, subject=subject, message=message,
                             html=html, files=files)

EMAIL_TEMPLATE = \
"""
<html>
<head>
    <link href="https://fonts.googleapis.com/css?family=Muli::100,200,300,400,500,600,700,800" rel="stylesheet">
</head>
    <body style="position: relative; float: left; width: 100%; height: 100%;  text-align: center; font-family: 'Muli', sans-serif;">
        <h1 style="float: left; width: 100%; margin: 20px 0px; font-size: 28px; text-align: center; color: #555555;">{title}</h1>
                <h2 style="float: left; width: 100%; margin: 40px 0px 10px 0px; font-size: 16px; text-align: center; color: #555555;">Activation Key:</h2>
        <h2 style="float: left; width: 100%; margin: 0px 0px 40px 0px; font-size: 24px; text-align: center; color: #61C0DF; font-weight: bold;">{ukey}</h2>
        <a  style="display: inline-block; width: 200px; height: 35px; margin: 20px auto 50px auto; padding-top: 15px; border-radius: 10px;
        background: #61C0DF; text-align: center; color: #FFFFFF; text-decoration: none;" href="{href}">SET PASSWORD</a>
    </body>
</html>
"""

def send_ukey_by_email(notify, email, host, ukey, product):
    try:
        subject = f'{product} set password'
        title = f'{product} set password'
        href = f'http://{host}/login/set_passwd?ukey={ukey}'
        html = EMAIL_TEMPLATE
        html = re.sub(r'{title}', title, html)
        html = re.sub(r'{ukey}', f'{ukey}', html)
        html = re.sub(r'{href}', f'{href}', html)
        notify.send_mail_async(email, subject=subject, html=html)
    except Exception as e:
        return str(e)
    return None


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


from distutils.core import setup

# setup(name='utils',
#       version='1.0.13',
#       description='utils lib',
#       author='João Neto',
#       author_email='joao.filipe.neto@gmail.com',
#       packages=['utils'],
# )

setup(name='dl2050utils',
      packages=['dl2050utils'],
      version='1.0.13',
      license='MIT',
      description='Utils lib',
      author='João Neto',
      author_email='joao.filipe.neto@gmail.com',
      keywords=['utils'],
      url='https://github.com/jn2050/utils',
      download_url='https://github.com/jn2050/utils/archive/v_1_0_13.tar.gz',
      # install_requires=[
      #       'pathlib',
      #       'zipfile',
      #       'json',
      #       'socket',
      #       'smtplib',
      #       'ssl',
      #       'boto3',
      #       'asyncpg',
      #       '',
      # ],
      classifiers=[
            'Development Status :: 4 - Beta',      # "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.7',
      ],
)
